/*

a:


b:


 */


#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "wrap/thread.h"
#include "wrap/synch.h"
#include "wrap/atomics.h"
#include "wrap/timer.h"

struct event;
struct animations;

// Typ för att lagra funktionspekare. Lagrar pekare till funktioner
// som ser ut som följer:
//   void handler(struct animations *, struct event *)
typedef void (*event_handler_fn)(struct animations *, struct event *);


/**
 * Datastruktur som representerar något som ska hända under en
 * animation. Lagrar en pekare till en funktion som implementerar
 * händelsen tillsammans med ett heltal som lagrar godtycklig data.
 */
struct event {
    // Funktion som ska anropas.
    event_handler_fn handler;

    // Data som funktionen kan använda.
    int data;
};


/**
 * Datastruktur som representerar pågående animationer i form av en
 * samling händelser (event). Alla händelser kan sedan köras med
 * funktionen 'animations_run_step'.
 *
 * Du kan anta att innehållet i datastrukturen bara används av
 * funktionerna som börjar med 'animations_' här nedanför. Det är
 * alltså okej att ersätta variabler här med synkroniseringsprimitiver
 * vid behov.
 */
struct animations {
    // Array av händelser.
    struct event *events;
    // Antal platser i 'events' som används just nu.
    int event_count;
    // Totalt antal platser som finns i 'events'.
    int event_max;

    // Antal trådar som är redo att hantera de händelser som
    // finns i 'events'.
    int thread_count;
    // Vilken händelse är nästa som trådarna ska behandla?
    int next_event;
    // Används för att meddela trådarna att det är dags att arbeta.
    bool start_work;
    // Används för att informera trådarna om att det är dags att
    // avslutas.
    bool exit;
};


// Funktion som körs av trådarna som startas av 'animations_create'.
// Trådarna väntar på att få information om att det är dags att arbeta
static void animations_worker(struct animations *ani) {
    bool should_exit = false;
    do {
        // Är det dags att göra något?
        while (!ani->start_work)
            ;

        // Ska vi stänga av oss?
        should_exit = ani->exit;

        while (!should_exit) {
            // Finns det något att göra?
            if (ani->next_event >= ani->event_count) {
                break;
            }
            // Plocka en händelse att behandla.
            int my_event = ani->next_event++;

            // Kör funktionen i händelsen.
            struct event *e = &ani->events[my_event];
            (*e->handler)(ani, e);
        }

    } while (!should_exit);
}


// Skapa ett animations-objekt med plats för maximalt 'event_max'
// händelser. Startar också 'thread_count' trådar för att hantera
// händelserna parallellt när 'animations_run_step' anropas.
struct animations *animations_create(int events_max, int thread_count) {
    struct animations *ani = malloc(sizeof(struct animations));
    ani->events = malloc(sizeof(struct event) * events_max);
    ani->event_count = 0;
    ani->event_max = events_max;
    ani->thread_count = thread_count;
    ani->next_event = 0;
    ani->start_work = false;
    ani->exit = false;

    // Starta trådar så att de är redo.
    for (int i = 0; i < thread_count; i++)
        thread_new(&animations_worker, ani);

    return ani;
}

// Lägg till en händelse till ett animationsobjekt. Detta ska kunna
// göras när som helst så länge animationsobjektet inte förstörs.
bool animations_add(struct animations *ani, struct event event) {
    if (ani->event_count >= ani->event_max)
        return false;

    ani->events[ani->event_count++] = event;

    return true;
}

// Kör alla händelser som finns i animationsobjektet och vänta tills
// alla händelser har körts klart. Du kan anta att
// 'animations_run_step' bara används av en tråd. Olika instanser ska
// dock kunna användas parallellt.
void animations_run_step(struct animations *ani) {
    ani->next_event = 0;
    ani->start_work = true;

    while (ani->next_event < ani->event_count)
        ;

    ani->start_work = false;
}

// Förstör ett animationsobjekt. Avslutar de trådar som skapades
// i 'animations_create' och avallokerar minnet för datastrukturen.
// Du kan anta att ingen annan tråd använder animationsobjektet
// när 'animations_destroy' anropas.
void animations_destroy(struct animations *ani) {
    // Be trådarna att avsluta sig själva, och väck dem sedan
    // genom att sätta 'start_work' till 'true'.
    ani->exit = true;
    ani->start_work = true;

    // Frigör minnet.
    free(ani->events);
    free(ani);
}

/**
 * Huvudprogram.
 *
 * Programmet finns för att visa hur koden ovan kan köras. Den är inte
 * nödvändigtvis designad för att visa alla möjliga problem i
 * implementationen.
 *
 * Ändringar nedanför denna kommentar kommer att ignoreras vid
 * bedömning av din lösning. Uppgiften ska alltså lösas utan att lägga
 * till synkronisering här nedanför. Det går självklart bra att
 * modifiera koden nedanför för att testa din lösning.
 */

// Lås för stdout.
struct lock stdout_lock;

// Händelse som skriver ut ett tecken till stdout.
void print_handler(struct animations *a, struct event *e) {
    (void)a;
    lock_acquire(&stdout_lock);
    printf("%c", (char)e->data);
    lock_release(&stdout_lock);
}

// Skapa en händelse som anropar 'print_handler' ovan och skriver ut
// ett tecken till skärmen.
void add_print_handler(struct animations *to, char to_print) {
    struct event event = {
        print_handler,
        (int)to_print
    };
    animations_add(to, event);
}

// Händelse som lägger till en ny händelse varje steg.
void create_handler(struct animations *a, struct event *e) {
    add_print_handler(a, (char)(e->data++));
}

// Skapa en händelse som lägger till nya händelser varje gång.
void add_create_handler(struct animations *to, char start) {
    struct event event = {
        create_handler,
        (int)start
    };
    animations_add(to, event);
}

// Main, använder 'animations' för att köra händelser i 10 steg.
int main(void) {
    const char *divider = "--------------------";
    lock_init(&stdout_lock);

    // Skapa en animation.
    struct animations *ani = animations_create(40, 4);

    // Lägg till en händelse som lägger till nya utskrifter av tecken varje iteration.
    add_create_handler(ani, 'A');

    // Lägg till några vanliga utskrifter till att börja med.
    for (int i = 0; i < 10; i++) {
        add_print_handler(ani, 'a' + i);
    }

    // Kör animationen i 10 steg.
    printf("%s\n", divider);
    for (int i = 0; i < 10; i++) {
        animations_run_step(ani);
        printf("\n%s\n", divider);
    }

    // Städa upp.
    animations_destroy(ani);

    return 0;
}
